@@ -176,7 +176,7 @@ void __connman_element_list(struct connman_element *element,
 						append_path, &filter);
 }
 
-static struct connman_network *__connman_element_get_network(struct connman_element *element)
+static struct connman_network *get_network(struct connman_element *element)
 {
 	if (element->type == CONNMAN_ELEMENT_TYPE_NETWORK &&
 						element->network != NULL)
@@ -185,7 +185,7 @@ static struct connman_network *__connman_element_get_network(struct connman_elem
 	if (element->parent == NULL)
 		return NULL;
 
-	return __connman_element_get_network(element->parent);
+	return get_network(element->parent);
 }
 
 struct connman_service *__connman_element_get_service(struct connman_element *element)
@@ -211,7 +211,7 @@ struct connman_service *__connman_element_get_service(struct connman_element *el
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-		network = __connman_element_get_network(element);
+		network = get_network(element);
 		if (network == NULL)
 			return NULL;
 		service = __connman_service_lookup_from_network(network);
@@ -259,6 +259,7 @@ const char *__connman_element_get_network_path(struct connman_element *element)
 struct find_data {
 	enum connman_service_type type;
 	struct connman_device *device;
+	connman_bool_t error;
 };
 
 static gboolean find_device(GNode *node, gpointer user_data)
@@ -341,6 +342,7 @@ static gboolean enable_technology(GNode *node, gpointer user_data)
 	struct connman_element *element = node->data;
 	struct find_data *data = user_data;
 	enum connman_service_type type;
+	int err;
 
 	if (element->type != CONNMAN_ELEMENT_TYPE_DEVICE)
 		return FALSE;
@@ -367,18 +369,23 @@ static gboolean enable_technology(GNode *node, gpointer user_data)
 		break;
 	}
 
-	__connman_device_enable_persistent(element->device);
+	err = __connman_device_enable_persistent(element->device);
+	if (err == 0 || (err < 0 && err == -EINPROGRESS))
+		data->error = FALSE;
 
 	return FALSE;
 }
 
 int __connman_element_enable_technology(enum connman_service_type type)
 {
-	struct find_data data = { .type = type, .device = NULL };
+	struct find_data data = { .type = type, .device = NULL, .error = TRUE };
 
 	g_node_traverse(element_root, G_PRE_ORDER,
 				G_TRAVERSE_ALL, -1, enable_technology, &data);
 
+	if (data.error == TRUE)
+		return -ENODEV;
+
 	return 0;
 }
 
@@ -387,6 +394,7 @@ static gboolean disable_technology(GNode *node, gpointer user_data)
 	struct connman_element *element = node->data;
 	struct find_data *data = user_data;
 	enum connman_service_type type;
+	int err;
 
 	if (element->type != CONNMAN_ELEMENT_TYPE_DEVICE)
 		return FALSE;
@@ -413,18 +421,23 @@ static gboolean disable_technology(GNode *node, gpointer user_data)
 		break;
 	}
 
-	__connman_device_disable_persistent(element->device);
+	err = __connman_device_disable_persistent(element->device);
+	if (err == 0 || (err < 0 && err == -EINPROGRESS))
+		data->error = FALSE;
 
 	return FALSE;
 }
 
 int __connman_element_disable_technology(enum connman_service_type type)
 {
-	struct find_data data = { .type = type, .device = NULL };
+	struct find_data data = { .type = type, .device = NULL, .error = TRUE };
 
 	g_node_traverse(element_root, G_PRE_ORDER,
 				G_TRAVERSE_ALL, -1, disable_technology, &data);
 
+	if (data.error == TRUE)
+		return -ENODEV;
+
 	return 0;
 }
 
@@ -642,6 +655,8 @@ void connman_element_unref(struct connman_element *element)
 			element->destruct(element);
 		free_children(element);
 		free_properties(element);
+		g_free(element->hostname);
+		g_free(element->domainname);
 		g_free(element->ipv4.address);
 		g_free(element->ipv4.netmask);
 		g_free(element->ipv4.gateway);
@@ -650,6 +665,8 @@ void connman_element_unref(struct connman_element *element)
 		g_free(element->ipv4.nameserver);
 		g_free(element->ipv4.timeserver);
 		g_free(element->ipv4.pac);
+		g_free(element->ipv6.address);
+		g_free(element->ipv6.network);
 		g_free(element->devname);
 		g_free(element->path);
 		g_free(element->name);
@@ -736,6 +753,18 @@ int connman_element_get_value(struct connman_element *element,
 		return -EINVAL;
 
 	switch (id) {
+	case CONNMAN_PROPERTY_ID_HOSTNAME:
+		if (element->hostname == NULL)
+			return connman_element_get_value(element->parent,
+								id, value);
+		*((char **) value) = element->hostname;
+		break;
+	case CONNMAN_PROPERTY_ID_DOMAINNAME:
+		if (element->domainname == NULL)
+			return connman_element_get_value(element->parent,
+								id, value);
+		*((char **) value) = element->domainname;
+		break;
 	case CONNMAN_PROPERTY_ID_IPV4_METHOD:
 		if (element->ipv4.method == CONNMAN_IPCONFIG_METHOD_UNKNOWN)
 			return connman_element_get_value(element->parent,
@@ -784,6 +813,13 @@ int connman_element_get_value(struct connman_element *element,
 								id, value);
 		*((char **) value) = element->ipv4.pac;
 		break;
+	case CONNMAN_PROPERTY_ID_IPV6_GATEWAY:
+		if (element->ipv6.gateway == NULL)
+			return connman_element_get_value(element->parent,
+								id, value);
+		*((char **) value) = element->ipv6.gateway;
+		break;
+
 	default:
 		return -EINVAL;
 	}
@@ -1349,6 +1385,7 @@ void __connman_element_start(void)
 	__connman_connection_init();
 	__connman_ipv4_init();
 	__connman_dhcp_init();
+	__connman_wpad_init();
 
 	__connman_rfkill_init();
 }
@@ -1359,6 +1396,7 @@ void __connman_element_stop(void)
 
 	__connman_rfkill_cleanup();
 
+	__connman_wpad_cleanup();
 	__connman_dhcp_cleanup();
 	__connman_ipv4_cleanup();
 	__connman_provider_cleanup();
